#include "imagereadingthread.h"
#include "imagereadevent.h"
#include <QtDebug>
#include <tango.h>
#include <macros.h>
#include <QTimer>
#include <QReadLocker>
#include <QWriteLocker>
#include <TLog>
#include <QCoreApplication>
#include <math.h>

ImageReadingThread::ImageReadingThread(QObject *parent,
                                       const QString &devname) :
    QThread(parent)
{
    mPeriod = 100;
    mImageDepth = 8;
    mProcessTimer = NULL;
    mDevice = NULL;
    mDevname = devname;
    try
    {
        mDevice = new Tango::DeviceProxy(mDevname.toStdString().c_str());
    }
    catch(Tango::DevFailed& e)
    {
        TLog l(e);
        ImageReadEvent *ire = new ImageReadEvent();
        ire->error = true;
        ire->message = l.formatMessage();
        ire->quality = Tango::ATTR_INVALID;
        qApp->postEvent(parent, ire);
    }
}

ImageReadingThread::~ImageReadingThread()
{
    printf("~ImageReadingThread()\n");
}

void ImageReadingThread::startPolling()
{
    if(!isRunning())
        this->start();
}

void ImageReadingThread::stopPolling()
{
    if(isRunning())
        emit exitLoop();
}

void ImageReadingThread::run()
{
    if(mProcessTimer)
        delete mProcessTimer;
    if(mDevice)
        delete mDevice;
    try
    {
        mDevice = new Tango::DeviceProxy(mDevname.toStdString().c_str());


        mProcessTimer = new QTimer(0);

        QTimer *exitTimer = new QTimer(0);
        QTimer *changePeriodTimer = new QTimer(0);

        connect(this, SIGNAL(changePeriod()), changePeriodTimer, SIGNAL(timeout()), Qt::QueuedConnection);
        connect(changePeriodTimer, SIGNAL(timeout()), this, SLOT(applyPeriodChange()), Qt::DirectConnection);

        connect(this, SIGNAL(exitLoop()), exitTimer, SIGNAL(timeout()), Qt::QueuedConnection);
        connect(exitTimer, SIGNAL(timeout()), this, SLOT(quit()), Qt::DirectConnection);

        connect(mProcessTimer, SIGNAL(timeout()), this, SLOT(read()), Qt::DirectConnection);

        /* event loop */
        qDebug() << this <<  QThread::currentThread() << __FUNCTION__ << "entering main loop";

        mProcessTimer->setSingleShot(false);
        mProcessTimer->setInterval(mPeriod);
        mProcessTimer->start();
        exec();

        qDebug() << this <<  QThread::currentThread() << __FUNCTION__ << "leaving main loop: deleting timers";

        delete exitTimer;
        delete changePeriodTimer;
    }
    catch(Tango::DevFailed& e)
    {
        TLog l(e);
        ImageReadEvent *ire = new ImageReadEvent();
        ire->error = true;
        ire->message = l.formatMessage();
        ire->quality = Tango::ATTR_INVALID;
        qApp->postEvent(parent(), ire);

    }
}

void ImageReadingThread::applyPeriodChange()
{
    /* this is called inside the thread */
    qDebug() << this <<  QThread::currentThread() << __FUNCTION__;
    mProcessTimer->setInterval(mPeriod);
}

void ImageReadingThread::setImageDepth(int dpt)
{
    QWriteLocker wl(&mRWLock);
    mImageDepth = dpt;
}

void ImageReadingThread::read()
{
    // qDebug() << this <<  QThread::currentThread() << __FUNCTION__ << QString("Image%1").arg(mImageDepth) << parent();
    ImageReadEvent *ire = new ImageReadEvent();
    QReadLocker rl(&mRWLock);
    DeviceAttribute imageAttr;
    try{
        imageAttr = mDevice->read_attribute(QString("Image%1").arg(mImageDepth).toStdString().c_str());
        ire->quality = imageAttr.get_quality();
        if (ire->quality != Tango::ATTR_INVALID)
        {
            ire->dimX = imageAttr.get_dim_x();
            ire->dimY = imageAttr.get_dim_y();
            ire->imageDepth = mImageDepth;
            if (mImageDepth == 16)
            {
                imageAttr >> ire->imageArray16;
            }
            else
            {
                imageAttr >> ire->imageArray8;
            }
        }
        else
        {
            ire->error = true;
            ire->message = QString("Image%1").arg(mImageDepth) + " attribute quality is invalid";
        }
    }
    catch(Tango::DevFailed &e)
    {
        TLog l(e);
        ire->error = true;
        ire->message = l.formatMessage();
        ire->quality = Tango::ATTR_INVALID;
    }
    /* notify */
    qApp->postEvent(parent(), ire);
}

void ImageReadingThread::setPeriod(int period)
{
    qDebug() << this <<  QThread::currentThread() << __FUNCTION__;
    mPeriod = period;
    emit changePeriod();
}
