#include "imagewidgetbase.h"
#include <QPaintEvent>
#include <QGLWidget>
#include <QPainter>
#include <QMouseEvent>
#include <QtDebug>
#include <macros.h>
#include "imagemouseeventinterface.h"
#include <externalscalewidget.h>
#include <math.h>
#include <elettracolors.h>
#include <QSettings>


ImageWidgetBase::ImageWidgetBase(QWidget *widget, bool isOpenGL)
{
    mWidget = widget;
    mIsOpenGL = isOpenGL;
    mError = false;
    mGridEnabled = false;
    mDiv = 16;
    mMin = 24;
    mMax = 40;
    mXScale = mYScale = NULL;
    mGridColor = QColor(Qt::white);
    mGridTextColor = QColor(Qt::white);
    mMouseEventIf = NULL;
    mZoom = 100;
}

ImageWidgetBase::~ImageWidgetBase()
{

}

void ImageWidgetBase::setErrorImageEnabled(bool en)
{
    QSettings s;
    s.setValue("PLAY_DOWN_ERRORS", en);
    if(en)
    {
        if(s.value("PLAY_DOWN_IMAGE", "okay").toString() == "okay")
            mErrorImage = QImage(":images/okguy.jpg");
        else if(s.value("PLAY_DOWN_IMAGE").toString() == "troll")
            mErrorImage = QImage(":images/trollface-problem.jpg");
    }
    else
        mErrorImage = QImage();
}

void ImageWidgetBase::initGridFromScales(ExternalScaleWidget *xScale, ExternalScaleWidget *yScale)
{
    mXScale = xScale;
    mYScale = yScale;
}

void ImageWidgetBase::mouseMove(QMouseEvent *ev)
{
    if(!mP1.isNull())
        mP2= ev->pos();
    if(mMouseEventIf)
        mMouseEventIf->imageMouseMoveEvent(ev->pos(), ev->button());
}

void ImageWidgetBase::mousePress(QMouseEvent *ev)
{
    qDebug() << __FUNCTION__ << ev->pos();
    mP1 = ev->pos();
    mLeftButton = (ev->button() == Qt::LeftButton);
    if(mMouseEventIf)
        mMouseEventIf->imageMousePressEvent(ev->pos(), ev->button());
}

void ImageWidgetBase::mouseRelease(QMouseEvent *ev)
{
    qDebug() << __FUNCTION__ << ev->pos() << mWidget->geometry();
    /* call the release method only if the button was pressed inside the label
     */
    QRect r(QPoint(0,0), mWidget->geometry().size());
    if(mMouseEventIf && !mP1.isNull() && r.contains(ev->pos()))
        mMouseEventIf->imageMouseReleaseEvent(ev->pos(), ev->button());
    else
        printf("\e[1;31m outside boundaries!@!\e[0m\n");

    mP1 = QPoint();
    mP2 = QPoint();
}

void ImageWidgetBase::setImageMouseEventInterface(ImageMouseEventInterface* ifa)
{
    mMouseEventIf = ifa;
}

void ImageWidgetBase::paint(QPaintEvent *e, QWidget *paintDevice)
{
    QPainter p(paintDevice);
    p.setClipRect(e->rect());

    p.drawImage(0, 0, mImage);
    if(!mP1.isNull() && !mP2.isNull())
    {
        QPen pen = p.pen();
        pen.setWidthF(2.0);
        QRect r(QPoint(0,0), mWidget->geometry().size());
        if(!r.contains(mP2))
            pen.setColor(Qt::red);
        else
            pen.setColor(Qt::green);

        p.setPen(pen);

        int p1x = mP1.x() * 100 / mZoom;
        int p1y = mP1.y() * 100 / mZoom;
        int p2x = mP2.x() * 100 / mZoom;
        int p2y = mP2.y() * 100 / mZoom;

        QFont f = p.font();
        f.setBold(true);
        p.setFont(f);
        QFontMetrics fm(p.font());

        QColor bg = Qt::white;
        bg.setAlpha(184);

        if(!mLeftButton)
        {
            float d = sqrt(pow(p2x - p1x, 2) + pow(p2y - p1y, 2));
            QString val = QString("%1").arg(d, 0, 'f', 2);
            p.fillRect(QRectF(mP1.x(), mP1.y(), fm.width(val), fm.height()), bg);
            /* first draw line */
            p.drawLine(mP1, mP2);
            /* then text */
            pen.setColor(Qt::black);
            p.setPen(pen);
            p.drawText(QPointF(mP1.x(), mP1.y() + fm.height()), val);
        }
        else
        {
            p.drawRect(QRect(mP1, mP2));
            p.setPen(bg);
            p.fillRect(QRect(mP1, mP2), bg);
            /* then text */
            QString val = QString("%1,%2 %3x%4").arg(p1x).arg(p1y).arg(fabs(p2x - p1x)).
                    arg(fabs(p2y - p1y));
            p.fillRect(QRectF(mP1.x(), mP1.y(), fm.width(val), fm.height()), bg);
            pen.setColor(Qt::black);
            p.setPen(pen);
            p.drawText(QPointF(mP1.x(), mP1.y() + fm.height()), val);
        }
    }
    if(mError)
    {
        QPen errPen(KDARKRED);
        errPen.setWidth(5);
        p.setPen(errPen);
        /* draw a cross on the figure */
        if(mErrorImage.isNull())
        {
            p.drawLine(e->rect().topLeft(), e->rect().bottomRight());
            p.drawLine(e->rect().bottomLeft(), e->rect().topRight());
        }
        else /* yeah! draw ok guy! */
        {
            if(mErrorImage.size() != e->rect().size())
                mErrorImage = mErrorImage.scaled(e->rect().size(), Qt::KeepAspectRatio);
            p.drawImage(e->rect(), mErrorImage);
        }

        QRect txtRect = e->rect();
        QColor bgColor = Qt::white;
        bgColor.setAlpha(180);
        p.fillRect(txtRect, bgColor);
        /* draw the error message */
        QFont f = p.font();
        f.setPointSize(10);
        f.setBold(true);
        p.setFont(f);
        p.drawText(txtRect, mErrorMessage);
    }
    else if(mGridEnabled && mXScale && mYScale)
    {
        QFont f = p.font();
        f.setBold(true);
        p.setFont(f);
        QRect r = mWidget->geometry();
        int w = r.width();
        int h = r.height();
        double x = 0;
        double val, tick;
        QString valueStr;
        QPen txtPen = p.pen();
        double xinterval = mXScale->maximum() - mXScale->minimum();
        double xdiv = (mXScale->maximum() - mXScale->minimum()) / mXScale->tickStepLen();
        double pixstep = (w - 1) / (double)xdiv; /* same calculations as in scale.cpp */
        /* draw vertical lines and labels */
        for(int i = 0; i < xdiv; i++)
        {
            tick = x + i * pixstep;
            p.setPen(mGridColor);
            p.drawLine(tick, h, tick, 0);

            if(mXScale->isInverted())
                val = mXScale->maximum() - (tick * xinterval / (double)w);
            else
                val =  mXScale->minimum() + tick * xinterval / (double)w;
            valueStr = QString("%1").arg(val, 0, 'f', 1);
            txtPen.setWidth(4.0);
            txtPen.setColor(mGridTextColor);
            p.setPen(txtPen);
            p.drawText(tick, h / 2, valueStr);
        }
        /* draw horizontal lines and labels */
        double ydiv = (mYScale->maximum() - mYScale->minimum()) / mYScale->tickStepLen();
        double yinterval = mYScale->maximum() - mYScale->minimum();
        /* recalculate pixstep according to height */
        pixstep = h / (double) ydiv;
        for(int i = 0; i < ydiv; i++)
        {
            tick = x + i * pixstep;
            p.setPen(mGridColor);
            p.drawLine(0, tick, 0 + r.width(), tick);

            if(!mYScale->isInverted())
               val = mYScale->maximum() - (tick * yinterval / (double)h);
            else
                val =  mYScale->minimum() + tick * yinterval / (double)h;
            valueStr = QString("%1").arg(val, 0, 'f', 1);
            txtPen.setWidth(4.0);
            txtPen.setColor(mGridTextColor);
            p.setPen(txtPen);
            p.drawText(0 + r.width()/2,  tick, valueStr);
        }

    }
}
