#include "circulardatamanager.h"
#include <QTWatcher>
#include <elettracolors.h>
#include <plotscenewidget.h>
#include <scenecurve.h>
#include <curveitem.h>
#include <linepainter.h>
#include <markeritem.h>
#include <QtDebug>
#include <QtAlgorithms>
#include <macros.h>

CircularDataManager::CircularDataManager(PlotSceneWidget *plot) :
    QObject(plot),
    QTangoComProxyReader(this)
{
    mNewestIndex = 0;
    mOldestColor = KLIGHTGRAY;
    mNewestColor = Qt::black;
    mBufferSize = 10;
    mAlphaChannel = 255;
    mManualColors  = false;
    mPlot  = plot;
    initColorMap();
    connect(qtangoComHandle(), SIGNAL(attributeAutoConfigured(const TangoConfigurationParameters*)),
            this, SLOT(attributeConfigured(const TangoConfigurationParameters*)));
    connect(qtangoComHandle(), SIGNAL(newData(const TVariant&)), this, SLOT(refresh(const TVariant&)));
}

void CircularDataManager::initColorMap()
{
    if(mBufferSize < 1 || mManualColors)
        return;

    int nColors = mBufferSize;
    qreal stepR = (mNewestColor.redF() - mOldestColor.redF())/nColors;
    qreal stepG = (mNewestColor.greenF() - mOldestColor.greenF())/nColors;
    qreal stepB = (mNewestColor.blueF() - mOldestColor.blueF())/nColors;
    mColorMap.clear();

    mColorMap << mNewestColor;

    if(nColors > 1)
    {
        QColor color;
        for(int i = 1; i < nColors; i++)
        {
            qreal c1r, c1g, c1b, c2r, c2g, c2b;
            /* get color 1 and 2 floating point values */
            c1r = mColorMap.at(i - 1).redF();
            c1g = mColorMap.at(i - 1).greenF();
            c1b = mColorMap.at(i - 1).blueF();
            c2r = c1r - stepR;
            c2g = c1g - stepG;
            c2b = c1b - stepB;
            color.setRedF(c2r);
            color.setGreenF(c2g);
            color.setBlueF(c2b);
            color.setAlpha(mAlphaChannel);
            mColorMap << color;
        }
    }
}

void CircularDataManager::setColorList(const QList<QColor> colors)
{
    mManualColors = true;
    mColorMap = colors;
}

void CircularDataManager::setBufferSize(int s)
{
    mBufferSize = s;
    mNewestIndex = 0;
    for(int i = 0; i < mCurves.size(); i++)
    {
        SceneCurve *c = mCurves.at(i);
        delete c;
    }
    mCurves.clear();

    initColorMap();
}

void CircularDataManager::setPeriod(int p)
{
    QTangoComProxyReader::setPeriod(p);
    emit periodChanged(p);
}

void CircularDataManager::setToolTip(const QString& msg)
{
    mPlot->setToolTip(msg);
}

/** \brief initialize x data to be used at every refresh
 *
 */
void CircularDataManager::attributeConfigured(const TangoConfigurationParameters *cp)
{
    TVariant v = cp->currentValue();
    QVector<double> data = v.toDoubleVector();
    for(int i = 0; i < data.size(); i++)
        mXData << i;
}

void CircularDataManager::refresh(const TVariant &v)
{
    QVector <double> yData = v.toDoubleVector();
    if(mCurves.size() < mBufferSize)
    {
//        printf("\e[1;24mRe-populating with new curves bufsiz %d mCurvesSize %d mNewestIndex %d\e[0m\n",
//               mBufferSize, mCurves.size(), mNewestIndex);
        /* add a new curve and configure it */
        QString crvName = realSource() + "[" + QString::number(mCurves.size()) + "]";
        SceneCurve *curve = mPlot->addCurve(crvName);
        curve->setBufferSize(mXData.size());
   //     mPlot->addConfigurableObjects(crvName, curve);

        /* curves represented by lines */
        CurveItem *curveItem = new CurveItem(curve);
        mPlot->scene()->addItem(curveItem);
        curve->installCurveChangeListener(curveItem);
        LinePainter *linePainter = new LinePainter(curveItem);
        linePainter->setLineColor(mColorMap.at(mCurves.size()));
//        printf("\e[1;32m + \e[0m creating new curve \"%s\" with color %d\n", qstoc(curve->name()), mCurves.size());
        mCurves << curve;
        curve->setData(mXData, yData);

    }
    else
    {
        mNewestIndex = mNewestIndex % mCurves.size();
        /* update data on the oldest curve, in the position indicated by
         * mNewestIndex
         */
        SceneCurve * curve = mCurves.at(mNewestIndex);
        curve->setData(mXData, yData);
//        printf("\e[1;35m + \e[0m set data on \"%s\" newestIdx %d curves size %d\n", qstoc(curve->name()),
//               mNewestIndex, mCurves.size());

        /* update colors */
        int colorIndex = 0; /* 0: newest color in map */
        for(int i = mNewestIndex; i >= 0; i--) /* newest curves */
        {
            if(mColorMap.size() > colorIndex)
            {
                LinePainter* linePainter = static_cast<LinePainter*>(curve->curveItem()->itemPainter());
                linePainter->setLineColor(mColorMap.at(colorIndex));
            }
            else
                perr("CircularDataManager::refresh: color map size < %d", colorIndex);

//            printf("\e[0;35m + \e[0m curve \"%s\" updated with color %d i = %d mNewwset %d\n",
//                   qstoc(mCurves.at(colorIndex)->name()), colorIndex, i, mNewestIndex);
            colorIndex++;

        }
        /* from the end of the curve list to the position of the mNewestIndex, the
         * curves get older
         */
        for(unsigned int i = mCurves.size() - 1; i > mNewestIndex; i--)
        {
            if(mColorMap.size() > colorIndex)
                static_cast<LinePainter*>(mCurves.at(i)->curveItem()->itemPainter())->setLineColor(mColorMap.at(colorIndex));
            else
                perr("CircularDataManager::refresh: color map size < %d", colorIndex);

//            printf("\e[0;34m + \e[0m curve \"%s\" updated with color %d\n",
//                   qstoc(mCurves.at(colorIndex)->name()), colorIndex);
            colorIndex++;
        }
        mNewestIndex++;
    }
}
