/***************************************************************************
 *   $Id: spectrometer.cpp,v 1.21 2014-11-25 09:18:50 giacomo Exp $
 *                                                                         *
 *   Copyright (C) 2008 by Fabio Asnicar  fabio.asnicar@elettra.trieste.it *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "spectrometer.h"
#include "imagereadingthread.h"
#include "imagereadevent.h"
#include "imagewidgetbase.h"
#include "fileio.h"
#include "usage.h"
#include "image.h"
#include "imagegl.h"
#include "profiles/profiles.h"
#include "infodialog.h"

#include <QScrollBar>
#include <QTWatcher>
#include <QTWriter>
#include <QColorDialog>
#include <QDir>
#include <QTime>
#include <QSettings>
#include <readersgroup.h>
#include <spectrumwriter.h>
#include <propertyreader.h>

Spectrometer::Spectrometer(QWidget *parent) : QWidget(parent)
{
    if (QApplication::arguments().size() <= 6)
    {
        Usage usage(0);
        exit(1);
    }

    ui.setupUi(this);
    QSettings settings;
    QString cameraName;
    delay = 100;
    mNRows = maxX = 0;
    mNCols = maxY = 0;
    zoom = 100;
    zoomStep = 25;
    for (int i=0; i<4; i++) {
        mRoi.push_back(0);
    }
    for (int i=0; i<2; i++) {
        mLastpoint.push_back(-1);
    }
    mImage = NULL;
    mImageMatrix = NULL;
    mImageWidget = NULL;
    mImageGLWidget = NULL;

    /* create the layout that will host the image widget */
    QVBoxLayout *vblo = new QVBoxLayout(ui.image);
    vblo->setObjectName("ImageWidgetLayout");
    vblo->setMargin(0);
    vblo->setSpacing(0);


    /* initialize values before connections */
    ui.refresh_SpinBox->setValue(1000/delay);
    /* before the connect(ui.cbUseGL,.... ) below! */
    ui.cbUseGL->setChecked(settings.value("useOpenGL", false).toBool());
    /* play down image (okay guy ) */
    ui.cbPlayDownErrors->setChecked(settings.value("PLAY_DOWN_ERRORS", true).toBool());

    /* reads from QSettings the desired system (openGL or not) and
     * creates the right widget accordingly.
     * Creates connections too!
     */
    createImageWidget();

    QString scr_name = "";

    /* watchers and writers */
    QTWatcher *colorWatcher = new QTWatcher(this);
    colorWatcher->attach(this, SLOT(updatePlotLevel(int)));
#if VER_MAJ > 4 || (VER_MAJ == 4 && VER_MIN >3) || (VER_MAJ == 4 && VER_MIN >=3 && VER_FIX >= 4)
    colorWatcher->setToolTipsDisabled(true);
#endif
    colorWatcher->setSource("$1/PlotLevel");

    /* hor and ver sign */
    QTWatcher *horSignWatcher = new QTWatcher(this);
    horSignWatcher->setObjectName("HorSignWatcher");
    horSignWatcher->attach(&mSignHor);
    horSignWatcher->setSource("$1/SignHor");

    QTWatcher *verSignWatcher = new QTWatcher(this);
    verSignWatcher->setObjectName("VerSignWatcher");
    verSignWatcher->attach(&mSignVer);
    verSignWatcher->setSource("$1/SignVer");

    /* the following four watchers are used to update the
     * tlWavelength label
     */
    QTWatcher *wlenWatcher = new QTWatcher(this);
    wlenWatcher->attach(&mWavelen);
    wlenWatcher->setSource("$2/Wavelength");

    QTWatcher *wlenSpanWatcher = new QTWatcher(this);
    wlenSpanWatcher->attach(&mWavelenSpan);
    wlenSpanWatcher->setSource("$2/WavelengthSpan");

    QTWatcher *p2micronWatcher = new QTWatcher(this);
    p2micronWatcher->attach(&mPixel2Micron);
    p2micronWatcher->setPeriod(10000);
    p2micronWatcher->setSource("$1/Pixel2micron");


//    QTWatcher *spectrumPeakPosWatcher = new QTWatcher(this);
//    spectrumPeakPosWatcher->setObjectName("SpectrumPeakPosWatcher");
//    connect(spectrumPeakPosWatcher->qtangoComHandle(), SIGNAL(newData(TVariant)),
//            this, SLOT(calculateWavelength(TVariant)));
//    spectrumPeakPosWatcher->setSource("$1/SpectrumPeakPos_d");

    /*   SearchBackgroundParam readers group and writer */
    ReadersGroup *rg = new ReadersGroup(this);
    rg->addReader(ui.backgroudLevels_tLabel,0);
    rg->addReader(ui.derivativeThreshold_tLabel,1);
    rg->setSource("$1/SearchBackgroundParam");

    SpectrumWriter *sw = new SpectrumWriter(this);
    sw->addElement(ui.backgroudLevels_tDoubleSpinBox);
    sw->addElement(ui.derivativeThreshold_tDoubleSpinBox);
    sw->setTargets("$1/SearchBackgroundParam");

    PropertyReader *oYagOriginPropertyReader = new PropertyReader(qApp->arguments().at(1), this, QString(), true);
    connect(oYagOriginPropertyReader, SIGNAL(valueRead(QString,QStringList)), this,
            SLOT(oYagOriginPropertyReady(QString, QStringList)));
    oYagOriginPropertyReader->read("OYagOrigin");

    initWidgets();

    /* create connections */
    connect(ui.save_ToolButton,         SIGNAL(clicked()),         this, SLOT(save()));
    connect(ui.snapshot_ToolButton,     SIGNAL(clicked()),         this, SLOT(snapshot()));
    connect(ui.processing_TCheckBox,    SIGNAL(stateChanged(int)), this, SLOT(processingEnable()));
    connect(ui.showControls_ToolButton, SIGNAL(toggled(bool)),     ui.tabWidget, SLOT(setVisible(bool)));
    connect(ui.zoomIn_ToolButton,       SIGNAL(clicked()),         this, SLOT(zoomIn()));
    connect(ui.zoomOut_ToolButton,      SIGNAL(clicked()),         this, SLOT(zoomOut()));
    connect(ui.zoomOriginal_ToolButton, SIGNAL(clicked()),         this, SLOT(zoomOriginal()));
    connect(ui.zoomBestFit_ToolButton,  SIGNAL(clicked()),         this, SLOT(zoomBestFit()));
    connect(ui.color_Slider,            SIGNAL(valueChanged(int)), this, SLOT(setPlotLevel(int)));
    connect(ui.refresh_SpinBox,         SIGNAL(valueChanged(int)), this, SLOT(setDelay(int)));
    connect(ui.measuring_groupBox,      SIGNAL(clicked(bool)),     this, SLOT(measuringEnable(bool)));
    connect(ui.bw_radioButton,          SIGNAL(toggled(bool)),     this, SLOT(setColorMap(bool)));
    connect(ui.c_radioButton,           SIGNAL(toggled(bool)),     this, SLOT(setColorMap(bool)));
    connect(ui.jet_radioButton,         SIGNAL(toggled(bool)),     this, SLOT(setColorMap(bool)));
    connect(ui.hot_radioButton,         SIGNAL(toggled(bool)),     this, SLOT(setColorMap(bool)));
    connect(ui.pbInfo,                  SIGNAL(clicked()),        this, SLOT(showInfo()));

    /* profiles button */
    connect(ui.pbProfiles, SIGNAL(clicked()), this, SLOT(showProfiles()));

    /* connection for open GL switch */
    connect(ui.cbUseGL, SIGNAL(toggled(bool)), this, SLOT(toggleOpenGLWidget(bool)));

    /* set the scroll area associated to each scale */
    connect(ui.image_ScrollArea->horizontalScrollBar(), SIGNAL(valueChanged(int)), ui.xScale,
            SLOT(scrollChanged(int)));
    connect(ui.image_ScrollArea->verticalScrollBar(), SIGNAL(valueChanged(int)), ui.yScale,
            SLOT(scrollChanged(int)));

    connect(ui.image_ScrollArea->horizontalScrollBar(), SIGNAL(rangeChanged(int,int)), ui.xScale,
            SLOT(scrollRangeChanged(int,int)));
    connect(ui.image_ScrollArea->verticalScrollBar(), SIGNAL(rangeChanged(int,int)), ui.yScale,
            SLOT(scrollRangeChanged(int,int)));

    connect(ui.image_ScrollArea->horizontalScrollBar(), SIGNAL(valueChanged(int)), ui.image,
            SLOT(update()));
    connect(ui.image_ScrollArea->verticalScrollBar(), SIGNAL(valueChanged(int)), ui.image,
            SLOT(update()));

    connect(this, SIGNAL(imageSizeChanged(QSizeF)), ui.xScale, SLOT(areaChanged(QSizeF)));
    connect(this, SIGNAL(imageSizeChanged(QSizeF)), ui.yScale, SLOT(areaChanged(QSizeF)));

    tangodev   = QApplication::arguments()[1];
    cameraName = QApplication::arguments().last();
    ui.eAppButtonBpm->setCmdLine(QString("bpmspectrometer %1 %2").arg(tangodev).arg
                                 (qApp->arguments().at(6)));

    for (int k=3; k<QApplication::arguments().size(); k++) {
        QString a = QApplication::arguments()[k];
        if(a == "--help")
        {
            Usage usage(this);
            exit(0);
        }
        else if (a.startsWith("--refresh"))
        {
            delay = a.section('=',1,1).toInt();
        }
        else if (a.startsWith("--scr"))
        {
            /* it used to be read inside refreshSlow() */
            scr_name = a.section('=',1,1);
            QTWatcher *tdpScrInstrumentWatcher = new QTWatcher(this);
            tdpScrInstrumentWatcher->attach(ui.camera_Label, SLOT(setText(QString)));
            tdpScrInstrumentWatcher->setSource(scr_name + "/Instrument");
        }
    }
    if (scr_name == "")
        ui.camera_Label->setText(cameraName);

    /* The reading thread. It will be started when ImageDepth has been read
         * (see the watcher below)
         */
    ImageReadingThread *thread = new ImageReadingThread(this, tangodev);
    thread->setObjectName("ImageReadingThread");
    thread->setPeriod(delay);

    /* this watcher must be created after the reading thread */
    QTWatcher *imageDepthWatcher  = new QTWatcher(this);
#if VER_MAJ > 4 || (VER_MAJ == 4 && VER_MIN >3) || (VER_MAJ == 4 && VER_MIN >=3 && VER_FIX >= 4)
    //imageDepthWatcher->setToolTipsDisabled(true);
#endif
    imageDepthWatcher->setObjectName("ImageDepthWatcher");
    imageDepthWatcher->attach(this, SLOT(imageDepthChanged(int)));
    imageDepthWatcher->setSource("$1/ImageDepth");

    /* after thread creation because setDelay changes thread timer */
    setDelay(ui.refresh_SpinBox->value());

}

Spectrometer::~Spectrometer()
{
    ImageReadingThread *thread = findChild<ImageReadingThread *>("ImageReadingThread");
    printf("* stopping polling thread and exiting\n");
    thread->stopPolling();
    thread->wait();
}

void Spectrometer::createImageWidget()
{
    QSettings s;
    const bool useGL = s.value("useOpenGL", false).toBool();
    QVBoxLayout *lo = findChild<QVBoxLayout *>("ImageWidgetLayout");
    if(useGL)
        mImageGLWidget = new ImageGL(this);
    else
        mImageWidget = new Image(this);
    if(mImageGLWidget)
        lo->addWidget(mImageGLWidget);
    else
        lo->addWidget(mImageWidget);
    /* for the grid on/off */
    initImageWidget();
}

void Spectrometer::toggleOpenGLWidget(bool openGL)
{
    QSettings s;
    if(mImageWidget != NULL && !openGL)
        return;
    else if(mImageGLWidget != NULL && openGL)
        return;
    else if(mImageWidget && openGL)
    {
        QVBoxLayout *lo = findChild<QVBoxLayout *>("ImageWidgetLayout");
        lo->removeWidget(mImageWidget);
        delete mImageWidget;
        mImageWidget = NULL;
        mImageGLWidget = new ImageGL(this);
        lo->addWidget(mImageGLWidget);
    }
    else if(mImageGLWidget && !openGL)
    {
        QVBoxLayout *lo = findChild<QVBoxLayout *>("ImageWidgetLayout");
        lo->removeWidget(mImageGLWidget);
        delete mImageGLWidget;
        mImageGLWidget = NULL;
        mImageWidget = new Image(this);
        lo->addWidget(mImageWidget);
    }
    s.setValue("useOpenGL", openGL);
    /* for the grid on/off */
    initImageWidget();
}

//void Spectrometer::calculateWavelength(const TVariant &v)
//{
//    if(v.quality() == Tango::ATTR_INVALID)
//    {
//        ui.tlWavelength->setText("####");
//        ui.tlWavelength->setDisabled(true);
//        ui.tlWavelength->setToolTip(v.message());
//    }
//    else
//    {
//        double spectrumPeakPos = v.toDouble();
//        double wl = (spectrumPeakPos - 500) * mWavelenSpan / 1000.0 * mPixel2Micron + mWavelen;
//        ui.tlWavelength->setEnabled(true);
//        ui.tlWavelength->setToolTip(v.message().remove("\"SpectrumPeakPos\":"));
//        AttributeInfo ai = v.getAttributeInfo();
//        QString txt;
//        txt.sprintf(ai.format.c_str(), wl);
//        ui.tlWavelength->setText(txt);
//    }
//}

void Spectrometer::imageDepthChanged(int dpt)
{
    if(dpt != imageDepth)
    {
        imageDepth = dpt;
        ImageReadingThread *rt = findChild<ImageReadingThread *>("ImageReadingThread");
        rt->setImageDepth(dpt);
        if(!rt->isRunning())
            rt->startPolling();
    }
}

void Spectrometer::initWidgets()
{
    mLowerBoundPoint = QPoint();
    mUpperBoundPoint = QPoint(0,0);
    generateColorMap();
    setColorMap(true);
    ui.c_radioButton->setChecked(true);
    ui.image_ScrollArea->setWidget(ui.image);
    ui.image_ScrollArea->setFrameShape(QFrame::NoFrame);

    processingEnable();
    ui.tabWidget->setHidden(true);
    ui.image_ScrollArea->setMouseTracking(true);

    ui.calibration_Tab->setEnabled(false);

    ui.xScale->setOrientation(ScaleItem::Horizontal);
    ui.xScale->setInverted(true);
    ui.yScale->setOrientation(ScaleItem::Vertical);
    ui.yScale->setInverted(false);
    ui.yScale->setTickDrawingInverted(true);
    ui.yScale->setProperty("labelDistFromTick", 10);
    ui.yScale->setMinimum(0.0);
    QFont scaleFont = ui.xScale->font();
    scaleFont.setBold(true);
    ui.xScale->setFont(scaleFont);
    ui.yScale->setFont(scaleFont);
    ui.xScale->setFormat("%.2f");

    /* the scale is updated once per second, according to the Wavelength,
     * the WavelenghtSpan and the pixel2micron, which are refreshed by
     * 3 separate QTWatcher.
     * The scale is updated by updateXScale, if necessary.
     */
    QTimer *scaleUpdateTimer = new QTimer(this);
    connect(scaleUpdateTimer, SIGNAL(timeout()), this, SLOT(updateScales()));
    scaleUpdateTimer->setInterval(1000);
    scaleUpdateTimer->start();
}

void Spectrometer::oYagOriginPropertyReady(const QString &, const QStringList& props)
{
    mOYagOriginX = props.first().toDouble();
    updateScales();
}

void Spectrometer::updateScales()
{
    /* setup scale limits */
    double span, center;

    span = -mWavelenSpan * mPixel2Micron;
    center = mWavelen - mOYagOriginX  * mPixel2Micron / 1000 * mWavelenSpan;
    if(ui.xScale->minimum() != center - span / 2.0 || ui.xScale->maximum() != center + span / 2.0)
    {
        ui.xScale->setMinimum(center - span / 2.0);
        ui.xScale->setMaximum(center + span / 2.0);
        ui.xScale->setTickStepLen(span/10.0);
        ui.xScale->update();
    }

    /* minimum is 0 and has been previously set inside initWidgets() */
    if(ui.yScale->maximum() != mPixel2Micron)
    {
        ui.yScale->setMaximum(mPixel2Micron);
        ui.yScale->setTickStepLen(mPixel2Micron/10.0);
        ui.yScale->update();
    }
}

/**
 * Converts the image array to a 32-bit aligned matrix and creates the QImage
 */
void Spectrometer::showImage()
{
    QImage scaledImage;
    QWidget *imageWidget;
    mImageWidget != NULL ? imageWidget = mImageWidget : imageWidget = mImageGLWidget;
    // qDebug() << "Spectrometer::showImage  START";
    mImage->setColorTable(colorMap);

    if (zoom != 100)
    {
        int z = (int)((float)mImage->height()*((float)zoom)/100);
        scaledImage = mImage->scaledToHeight(z);
    }
    else
    {
        scaledImage = *mImage;
    }
    imageWidget->setGeometry(imageWidget->x(),
                             imageWidget->y(),
                             scaledImage.width(),
                             scaledImage.height());

    ui.image->resize(imageWidget->size());

    /* now we use QWidget + QPainter to draw!!!! */
    if(mImageWidget)
        mImageWidget->setImage(scaledImage);
    else
        mImageGLWidget->setImage(scaledImage);

    /* first time: center the image inside the scroll area */
    if(lastImageSize.isEmpty())
    {
        emit imageSizeChanged(scaledImage.size());
        QScrollBar *hScroll = ui.image_ScrollArea->horizontalScrollBar();
        QScrollBar *vScroll = ui.image_ScrollArea->verticalScrollBar();
        hScroll->setValue((hScroll->minimum() + hScroll->maximum())/2.0);
        vScroll->setValue((vScroll->minimum() + vScroll->maximum())/2.0);
    }
    if(lastImageSize != scaledImage.size())
    {
        lastImageSize = scaledImage.size();
        emit imageSizeChanged(lastImageSize);
    }
}

bool Spectrometer::event(QEvent *e)
{
    if(e->type() == QEvent::User + 1)
    {
        ImageWidgetBase *imageWidgetBase;
        mImageWidget != NULL ? imageWidgetBase = mImageWidget : imageWidgetBase = mImageGLWidget;

        ImageReadEvent *ire = static_cast<ImageReadEvent *>(e);
        imageWidgetBase->setError(ire->error);
        if(ire->error)
        {
            imageWidgetBase->setErrorMessage(ire->message);
            ui.image->update();
        }
        else
        {
            /* NOTE: x and y are switched: it is correct */
            if(ire->dimY != maxX || ire->dimX != maxY)
                initMatrix(ire->dimY, ire->dimX);

            if(ire->imageDepth == 16)
                vector2image16(ire->imageArray16);
            else
                vector2image8(ire->imageArray8);
            showImage();
        }
        /* event recognized: return true */
        return true;
    }
    else
        return QWidget::event(e);
}

void Spectrometer::initMatrix(int dimX, int dimY) {
    if(dimX != maxX || dimY != maxY)
    {

        if(mImageMatrix)
            delete mImageMatrix;

        /* update max dimensions and allocate a new matrix */
        mNRows = maxX  = dimX;
        maxY  = dimY;
        diffY = 4 - maxY % 4;
        mNCols = maxY + diffY;
        printf("\e[1;32mcreating matrix %d!\e[0m\n",  mNRows * mNCols);
        mImageMatrix = (unsigned char *) malloc(sizeof(unsigned char) * mNRows * mNCols);
    }
}

void Spectrometer::vector2image16(const std::vector<unsigned short> &image)
{
    uchar black = QString("%1").arg(0).data()->row();
    int x, y, index = 0;

    for(x = 0; x < maxX; x++)
    {
        for (y = 0; y < maxY; y++)
        {
            mImageMatrix[x * mNCols + y] = image[index] >> 4;
            /* save current image */
            index++;
        }
        for (y=0; y < diffY; y++)
        {
            mImageMatrix[x * mNCols + maxY + y] = black;
        }
    }
    // ATTENTION X & Y are switched: it is correct
    if(mImage)
        delete mImage;
    mImage = new QImage((const uchar *)mImageMatrix, maxY + diffY, maxX, QImage::Format_Indexed8);
}

void Spectrometer::vector2image8(const std::vector<unsigned char> & image)
{
    uchar black = QString("%1").arg(0).data()->row();
    int x, y, index = 0;

    /* directly save current image */
    for(x = 0; x < maxX; x++)
    {
        for (y = 0; y < maxY; y++)
        {
            mImageMatrix[(x * mNCols) + y] = image[index++];
        }
        for (y=0; y<diffY; y++)
        {
            mImageMatrix[(x * mNCols) + maxY+y] = black;
        }
    }
    if (mImage)
        delete mImage;
    mImage = new QImage((const uchar *)mImageMatrix, maxY + diffY, maxX, QImage::Format_Indexed8);
}

void Spectrometer::initImageWidget()
{
    QWidget *w = getImageWidget();
    connect(ui.cbGridOn, SIGNAL(toggled(bool)), w, SLOT(setGridEnabled(bool)));
    connect(ui.cbPlayDownErrors, SIGNAL(toggled(bool)), w, SLOT(setErrorImageEnabled(bool)));
    w->setCursor(QCursor(QPixmap(":images/mycross.xpm"),15,15));
    if(mImageGLWidget)
    {
        mImageGLWidget->setImageMouseEventInterface(this);
        mImageGLWidget->initGridFromScales(ui.xScale, ui.yScale);
        mImageGLWidget->setGridEnabled(ui.cbGridOn->isChecked());
        mImageGLWidget->setErrorImageEnabled(ui.cbPlayDownErrors->isChecked());
    }
    else
    {
        mImageWidget->setImageMouseEventInterface(this);
        mImageWidget->initGridFromScales(ui.xScale, ui.yScale);
        mImageWidget->setGridEnabled(ui.cbGridOn->isChecked());
        mImageWidget->setErrorImageEnabled(ui.cbPlayDownErrors->isChecked());
    }
}

QWidget *Spectrometer::getImageWidget() const
{
    if(mImageWidget)
        return mImageWidget;
    else
        return mImageGLWidget;
}

void Spectrometer::showProfiles()
{
    Profiles *profiles = findChild<Profiles *>();
    if(!profiles)
    {
        printf("- creating Profiles widget...\n");
        profiles = new Profiles(this);
        profiles->show();
    }
    else /* might have been minimized but not closed */
    {
        profiles->activateWindow();
    }
}

/*
 * =================================================================
 *
 * mouse move, press and release events. The following three methods
 * implement the ImageMouseEventInterface. The ImageMouseEventInterface
 * constitutes major news, algorithms have been checked but are taken
 * from the gigecam implementation.
 *
 * =================================================================
 */

void Spectrometer::imageMousePressEvent(const QPoint& pos, int )
{
    mRoi[0] = pos.x() * 100 / zoom;
    mRoi[1] = pos.y() * 100 / zoom;

    if (ui.measuring_groupBox->isChecked())
    {
        if (mLastpoint[0] != -1)
        {
            float d_px = sqrt(pow(mRoi[0] - mLastpoint[0],2) + pow(mRoi[1] - mLastpoint[1],2));
            ui.measureValue_px_Label->setText(QString("%1").arg(d_px, 0, 'f', 3));

            float d_mm = sqrt(pow((mRoi[0] - mLastpoint[0]) / ui.scaleHor_tReaderWriter->currentValue().toDouble(),2) +
                              pow((mRoi[1] - mLastpoint[1])/ui.scaleVer_tReaderWriter->currentValue().toDouble(),2));

            ui.measureValue_mm_Label->setText(QString("%1").arg(d_mm, 0, 'f', 3));
            ui.point_1_px_Label->setText(QString("%1,%2").arg(mLastpoint[0]).arg(mLastpoint[1]));
            ui.point_1_mm_Label->setText(QString("%1,%2").arg(normalize(mLastpoint[0], 'H'), 0, 'f', 3).arg(normalize(mLastpoint[1], 'V'), 0, 'f', 3));
        }
        mLastpoint[0] = mRoi[0];
        mLastpoint[1] = mRoi[1];
        ui.point_2_px_Label->setText(QString("%1,%2").arg(mLastpoint[0]).arg(mLastpoint[1]));
        ui.point_2_mm_Label->setText(QString("%1,%2").arg(normalize(mLastpoint[0], 'H'), 0, 'f', 3).arg(normalize(mLastpoint[1], 'V'), 0, 'f', 3));
    }
}

void Spectrometer::imageMouseReleaseEvent(const QPoint& pos, int )
{
    mRoi[2] = pos.x() * 100/zoom;
    mRoi[3] = pos.y() * 100/zoom;

    if(!ui.measuring_groupBox->isChecked())
    {
        printf("\e[1;35m controllare esecuzione ATTR roiParamWriter\e[0m (QTWriter)\n");
        QTWriter *roiParamWriter = new QTWriter(this);
        roiParamWriter->setAutoDestroy(true);
        roiParamWriter->setTargets("$1/RoiParam");
        roiParamWriter->execute(mRoi);
    }
}

void Spectrometer::imageMouseMoveEvent(const QPoint& pos, int )
{
    int x_px, y_px;
    x_px = pos.x() * 100/zoom;
    y_px = pos.y() * 100/zoom;
    ui.mouse_px_Label->setText(QString("%1,%2").arg(x_px).arg(y_px));
    ui.mouse_mm_Label->setText(QString("%1,%2").arg(normalize(x_px, 'H'), 0, 'f', 3).
                               arg(normalize(y_px, 'V'), 0, 'f', 3));
}

/*
 * =================================================================
 *
 * From here on, old code and some slot implementations
 * (some of them have been revised)
 *
 * =================================================================
 */

void Spectrometer::processingEnable()
{
    QList<QWidget *>readers = ui.measure_Frame->findChildren<QWidget *>();
    bool b = ui.processing_TCheckBox->isChecked();

    ui.timeValue_TLabel->setEnabled(b);
    ui.fastProcessing_TCheckBox->setEnabled(b);
    ui.autoRoi_TCheckBox->setEnabled(b);
    ui.roiThreshold_tReaderWriter->setEnabled(b);
    ui.acquireBackground_TPushButton->setEnabled(b);
    ui.bgSubtraction_TCheckBox->setEnabled(b);
    foreach(QWidget *w, readers)
    {
        if (qobject_cast<TLabel *>(w))
        {
            w->setEnabled(b);
        }
    }
}

QString Spectrometer::getFileName()
{
    QString tangodev = qApp->arguments().at(1);
    QString fileName(QDir::homePath());
    fileName.append("/CCD/");
    fileName.append(tangodev.right(tangodev.size() - tangodev.lastIndexOf("/") - 1));
    fileName.append("_");
    if (!ui.snapshotPrefix_LineEdit->text().isNull() && !ui.snapshotPrefix_LineEdit->text().isEmpty()) {
        fileName.append(ui.snapshotPrefix_LineEdit->text());
        fileName.append("_");
    }
    fileName.append(QTime::currentTime().toString("hhmmss"));
    // qDebug() << "Spectrometer::getFileName [" << QDir::homePath() << "][" << fileName << "]" ;
    return fileName;
}

/**
 * save the TIFF image
 */
void Spectrometer::snapshot()
{
    if(mImage)
    {
        /* immediately copy the current image
         * (it may be updated while choosing the file to save)
         */
        QImage snapImage = *mImage;
        FileIO fio;
        fio.snapshot(snapImage, getFileName());
    }
    else
        perr("Image null");
}

/**
 * save the image data
 */
void Spectrometer::save()
{
    if(mImage)
    {
        QImage snapImage = *mImage;
        FileIO fio;
        fio.save(snapImage.bits(), maxX, maxY, getFileName());
    }
}

/* this code has been taken from the refresh() method
 */
void Spectrometer::updatePlotLevel(int value)
{
    int plotLevel;
    if (imageDepth == 16)
        plotLevel = round(value/16);
    else
        plotLevel = value;

    if(colorMap.size() > plotLevel)
    {
        /* jointly update the slider */
        disconnect(ui.color_Slider, SIGNAL(valueChanged(int)), this, SLOT(setPlotLevel(int)));
        ui.color_Slider->setValue(plotLevel);
        connect(ui.color_Slider,    SIGNAL(valueChanged(int)), this, SLOT(setPlotLevel(int)));
        QPalette p = palette();
        p.setColor(QPalette::Button, colorMap[plotLevel]);
        ui.plotColor_PushButton->setPalette(p);
    }
    else
        perr("updatePlotLevel: color map size less than plot level");
}

void Spectrometer::setPlotLevel(int c)
{
    // qDebug() << "Spectrometer::setPlotLevel " << c;
    /* update push button color */
    QPalette p = palette();
    p.setColor(QPalette::Button, colorMap[c]);
    ui.plotColor_PushButton->setPalette(p);
    short newc;

    (imageDepth == 16) ? (newc = c*16) : (newc = c);

    QTWriter *writer = findChild<QTWriter *>("PlotLevelWriter");
    if(!writer)
    {
        writer = new QTWriter(this);
        writer->setObjectName("PlotLevelWriter");
        writer->setTargets("$1/PlotLevel");
    }
    writer->execute(newc);
}


float Spectrometer::normalize(float p, char axis)
{
    float res;
    if (axis == 'H')
        res = ((float)mSignHor) * -((((maxY/2) -p) /
                                     ui.scaleHor_tReaderWriter->currentValue().toDouble()) -
                                    ui.offsetHor_tReaderWriter->currentValue().toDouble());
    else
        res = ((float)mSignVer) *  ((((maxX/2) -p) /
                                     ui.scaleVer_tReaderWriter->currentValue().toDouble()) -
                                    ui.offsetVer_tReaderWriter->currentValue().toDouble());
    return res;
}

void Spectrometer::zoomIn()
{
    setZoom(zoom + zoomStep);
}

void Spectrometer::zoomOut()
{
    setZoom(zoom - zoomStep);
}

void Spectrometer::zoomOriginal()
{
    setZoom(100);
}

void Spectrometer::zoomBestFit()
{
    int zommH = (int)((float)(ui.image_ScrollArea->width()) * 100.0/mImage->width());
    int zoomV = (int)((float)(ui.image_ScrollArea->height()) * 100.0/mImage->height());
    setZoom(min(zommH,zoomV));
}

void Spectrometer::setZoom(int newZoom)
{
    ImageWidgetBase *imageWidgetBase;
    mImageWidget != NULL ? imageWidgetBase = mImageWidget : imageWidgetBase = mImageGLWidget;
    imageWidgetBase->setZoom(newZoom);

    zoom = newZoom;
    showImage();
    ui.zoom_Label->setText(QString("%1 %").arg(zoom));
    qDebug() << "areaChanged" << mImage->size() * zoom/100.0;
    ui.xScale->areaChanged(mImage->size() * zoom/100.0);
    ui.yScale->areaChanged(mImage->size() * zoom/100.0);
}

void Spectrometer::setDelay(int d)
{
    delay = 1000/d;

    findChild<ImageReadingThread *>("ImageReadingThread")->setPeriod(delay);
//    findChild<QTWatcher *>("SpectrumPeakPosWatcher")->setPeriod(delay);

    foreach(TLabel *label, ui.measure_Frame->findChildren<TLabel *>())
        label->setPeriod(delay);

    foreach(TLabel *label, ui.groupBoxSpectrumValues->findChildren<TLabel *>())
        label->setPeriod(delay);
    TUtil::instance()->setGlobalRefreshInterval(delay*0.9);
}

void Spectrometer::measuringEnable(bool enable)
{
    if (!enable)
    {
        ui.measureValue_px_Label->setText("");
        ui.measureValue_mm_Label->setText("");
        ui.point_1_px_Label->setText("");
        ui.point_1_mm_Label->setText("");
        ui.point_2_px_Label->setText("");
        ui.point_2_mm_Label->setText("");
        mLastpoint[0] = -1;
        mLastpoint[1] = -1;
        // qDebug() << "Spectrometer::measuringEnable  ClearMeasurePoints (" <<  mLastpoint[0] << "|" << mLastpoint[1] << ")";
    }
}

void Spectrometer::showInfo()
{
    InfoDialog id(this);
    id.exec();
}

